# -*- coding: utf-8 -*-

from __future__ import (print_function, unicode_literals, absolute_import,
                        division)
from pusher.config import Config
from pusher.request import Request
from pusher.sync import SynchronousBackend
from pusher.util import GET, POST, text, validate_channel

import collections
import json
import six

class RequestMethod(object):
    def __init__(self, pusher, f):
        self.pusher = pusher
        self.f = f

    def __call__(self, *args, **kwargs):
        return self.pusher.backend.send_request(self.make_request(*args, **kwargs))

    def make_request(self, *args, **kwargs):
        return self.f(self.pusher, *args, **kwargs)

def doc_string(doc):
    def decorator(f):
        f.__doc__ = doc
        return f
    return decorator

def request_method(f):
    @property
    @doc_string(f.__doc__)
    def wrapped(self):
        return RequestMethod(self, f)
    return wrapped

def join_attributes(attributes):
    for attr in attributes:
        if not isinstance(attr, six.text_type):
            raise TypeError('Each attr should be %s' % text)

    return six.text_type(',').join(attributes)

class Pusher(object):
    """Client for the Pusher HTTP API.

    This client supports various backend adapters to support various http
    libraries available in the python ecosystem. 

    :param config: a pusher.Config instance
    :param backend: an object that responds to the send_request(request)
                    method. If none is provided, a
                    python.sync.SynchronousBackend instance is created.
    """
    def __init__(self, config, backend=None):
        if not isinstance(config, Config):
            raise TypeError("config should be a pusher.Config object")
        self.backend = backend or SynchronousBackend(config)
        self.config = config

    @request_method
    def trigger(self, channels, event_name, data, socket_id=None):
        '''
        Trigger an event on one or more channels, see:

        http://pusher.com/docs/rest_api#method-post-event
        '''
        if isinstance(channels, six.string_types) or not isinstance(channels, (collections.Sized, collections.Iterable)):
            raise TypeError("Expected a collection of channels (each channel should be %s)" % text)

        if len(channels) > 10:
            raise ValueError("Too many channels")

        for channel in channels:
            validate_channel(channel)

        if not isinstance(event_name, six.text_type):
            raise TypeError("event_name should be %s" % text)

        if len(event_name) > 200:
            raise ValueError("event_name too long")

        if not isinstance(data, six.text_type):
            data = json.dumps(data)

        if len(data) > 10240:
            raise ValueError("Too much data")

        params = {
            'name': event_name,
            'channels': channels,
            'data': data
        }
        if socket_id:
            if not isinstance(socket_id, six.text_type):
                raise TypeError("Socket ID should be %s" % text)
            params['socket_id'] = socket_id
        return Request(self.config, POST, "/apps/%s/events" % self.config.app_id, params)

    @request_method
    def channels_info(self, prefix_filter=None, attributes=[]):
        '''
        Get information on multiple channels, see:

        http://pusher.com/docs/rest_api#method-get-channels
        '''
        params = {}
        if attributes:
            params['info'] = join_attributes(attributes)
        if prefix_filter:
            params['filter_by_prefix'] = prefix_filter
        return Request(self.config, GET, "/apps/%s/channels" % self.config.app_id, params)

    @request_method
    def channel_info(self, channel, attributes=[]):
        '''
        Get information on a specific channel, see:

        http://pusher.com/docs/rest_api#method-get-channel
        '''
        validate_channel(channel)

        params = {}
        if attributes:
            params['info'] = join_attributes(attributes)
        return Request(self.config, GET, "/apps/%s/channels/%s" % (self.config.app_id, channel), params)

    @request_method
    def users_info(self, channel):
        '''
        Fetch user ids currently subscribed to a presence channel

        http://pusher.com/docs/rest_api#method-get-users
        '''
        validate_channel(channel)

        return Request(self.config, GET, "/apps/%s/channels/%s/users" % (self.config.app_id, channel))



# -*- coding: utf-8 -*-

import aiohttp
import pusher

class AsyncIOBackend:
    """Adapter for the aiohttp module.

    This backend is only availble for python 3 users and doesn't support SSL.

    :param config: pusher.Config instance
    """
    def __init__(self, config):
        self.config = config
        if config.ssl:
            raise NotImplementedError("SSL not supported for this backend")

    def send_request(self, request):
        method = request.method
        url = "http://%s:%s%s" % (self.config.host, self.config.port, request.path)
        params = request.query_params
        data = request.body
        headers = request.headers

        response = yield from aiohttp.request(method, url, params=params, data=data, headers=headers)
        body = yield from response.read_and_close()
        return pusher.process_response(response.status, body.decode('utf8'))



# -*- coding: utf-8 -*-

from __future__ import (print_function, unicode_literals, absolute_import,
                        division)

import json
import re
import six
import sys

channel_name_re = re.compile('^[-a-zA-Z0-9_=@,.;]+$')
app_id_re       = re.compile('^[0-9]+$')

GET, POST, PUT, DELETE = "GET", "POST", "PUT", "DELETE"

if sys.version_info < (3,):
    text = 'a unicode string'
else:
    text = 'a string'

class PusherError(Exception):
    pass

class PusherBadRequest(PusherError):
    pass

class PusherBadAuth(PusherError):
    pass

class PusherForbidden(PusherError):
    pass

class PusherBadStatus(PusherError):
    pass

def process_response(status, body):
    if status == 200:
        return json.loads(body)
    elif status == 400:
        raise PusherBadRequest(body)
    elif status == 401:
        raise PusherBadAuth(body)
    elif status == 403:
        raise PusherForbidden(body)
    else:
        raise PusherBadStatus("%s: %s" % (status, body))

def validate_channel(channel):
    if not isinstance(channel, six.text_type):
        raise TypeError("Channel should be %s" % text)

    if len(channel) > 200:
        raise ValueError("Channel too long")

    if not channel_name_re.match(channel):
        raise ValueError("Invalid Channel: %s" % channel)



# -*- coding: utf-8 -*-

from __future__ import (print_function, unicode_literals, absolute_import,
                        division)

import tornado
import tornado.httpclient

class TornadoBackend(object):
    """Adapter for the tornado.httpclient module.

    :param config: a pusher.Config instance
    :param timeout: configurable timeout for the connection
    """
    def __init__(self, config, timeout=None):
        self.config = config
        self.timeout = timeout
        if self.config.ssl:
            raise NotImplementedError("SSL not supported for this backend")

    def send_request(self, request):
        method = request.method
        url = "http://%s:%s%s?%s" % (self.config.host, self.config.port, request.path, request.query_string)
        data = request.body
        headers = {'Content-Type': 'application/json'}

        return tornado.httpclient.HTTPRequest(url, method=method, body=data, request_timeout=self.timeout, headers=headers)



# -*- coding: utf-8 -*-

from __future__ import (print_function, unicode_literals, absolute_import,
                        division)
from pusher.util import process_response

import requests

class RequestsBackend(object):
    """Adapter for the requests module.

    :param config: an instance of pusher.Config
    :param options: key-value passed into the requests.request constructor
    """
    def __init__(self, config, **options):
        self.options = {'verify': True}
        self.options.update(options)

    def send_request(self, request):
        resp = requests.request(
            request.method,
            request.url,
            headers=request.headers,
            data=request.body,
            **self.options
        )
        return process_response(resp.status_code, resp.text)



# -*- coding: utf-8 -*-
from setuptools import setup
setup(
    name='pusher-rest',
    version='0.1.0',
    description='A Python library to interract with the Pusher API',
    url='http://github.com/pusher/pusher-rest-python',
    author='Pusher',
    author_email='support@pusher.com',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP",
    ],
    keywords='pusher rest realtime websockets service',
    license='MIT',

    packages=['pusher'],

    install_requires=['six'],
    tests_require=['nose', 'mock'],

    extras_require={
        'aiohttp': ["aiohttp>=0.9.0"],
        'requests': ['requests>=2.3.0'],
        'tornado': ['tornado>=4.0.0'],
    },

    test_suite='pusher_tests',
)



# -*- coding: utf-8 -*-

from __future__ import (print_function, unicode_literals, absolute_import,
                        division)
from pusher.util import app_id_re, channel_name_re, text

import hashlib
import hmac
import json
import os
import re
import six
import time

try:
    compare_digest = hmac.compare_digest
except AttributeError:
    # Not secure when the length is supposed to be kept secret
    def compare_digest(a, b):
        if len(a) != len(b):
            return False
        return reduce(lambda x, y: x | y, [ord(x) ^ ord(y) for x, y in zip(a, b)]) == 0

class Config(object):
    """The Config class holds the pusher credentials and other connection
    infos to the HTTP API.

    :param app_id: The Pusher application ID
    :param key: The Pusher application key
    :param secret: The Pusher application secret
    :param ssl: Whenever to use SSL or plain HTTP 
    :param host: Used for custom host destination
    :param port: Used for custom port destination
    :param cluster: Convention for other clusters than the main Pusher-one.
      Eg: 'eu' will resolve to the api-eu.pusherapp.com host

    Usage::

      >> from pusher import Config
      >> c = Config('455', 'mykey', 'mysecret')
    """
    def __init__(self, app_id, key, secret, ssl=False, host=None, port=None, cluster=None):
        if not isinstance(app_id, six.text_type):
            raise TypeError("App ID should be %s" % text)

        if not isinstance(key, six.text_type):
            raise TypeError("Key should be %s" % text)

        if not isinstance(secret, six.text_type):
            raise TypeError("Secret should be %s" % text)

        if not app_id_re.match(app_id):
            raise ValueError("Invalid app id")

        if port and not isinstance(port, six.integer_types):
            raise TypeError("Port should be a number")

        if not isinstance(ssl, bool):
            raise TypeError("SSL should be a boolean")

        self.app_id = app_id
        self.key = key
        self.secret = secret

        if host:
            if not isinstance(host, six.text_type):
                raise TypeError("Host should be %s" % text)

            self.host = host
        elif cluster:
            if not isinstance(cluster, six.text_type):
                raise TypeError("Cluster should be %s" % text)

            self.host = "api-%s.pusher.com" % cluster
        else:
            self.host = "api.pusherapp.com"

        self.port = port or (443 if ssl else 80)
        self.ssl = ssl

    @classmethod
    def from_url(cls, url):
        """Alternate constructor that extracts the information from a URL.

        :param url: String containing a URL

        Usage::

          >> from pusher import Config
          >> c = Config.from_url("http://mykey:mysecret@api.pusher.com/apps/432")
        """
        m = re.match("(http|https)://(.*):(.*)@(.*)/apps/([0-9]+)", url)
        if not m:
            raise Exception("Unparsable url: %s" % url)
        ssl = m.group(1) == 'https'
        return cls(key=m.group(2), secret=m.group(3), host=m.group(4), app_id=m.group(5), ssl=ssl)

    @classmethod
    def from_env(cls, env='PUSHER_URL'):
        """Alternate constructor that extracts the information from an URL
        stored in an environment variable. The pusher heroku addon will set
        the PUSHER_URL automatically when installed for example.

        :param env: Name of the environment variable

        Usage::

          >> from pusher import Config
          >> c = Config.from_env("PUSHER_URL")
        """
        val = os.environ.get(env)
        if not val:
            raise Exception("Environment variable %s not found" % env)
        return cls.from_url(six.text_type(val))

    @property
    def scheme(self):
        """Returns "http" or "https" scheme depending on the ssl setting."""
        return 'https' if self.ssl else 'http'

    def authenticate_subscription(self, channel, socket_id, custom_data=None):
        """Used to generate delegated client subscription token.

        :param channel: name of the channel to authorize subscription to
        :param socket_id: id of the socket that requires authorization
        :param custom_data: used on presence channels to provide user info
        """
        if not isinstance(channel, six.text_type):
            raise TypeError('Channel should be %s' % text)

        if not channel_name_re.match(channel):
            raise ValueError('Channel should be a valid channel, got: %s' % channel)

        if not isinstance(socket_id, six.text_type):
            raise TypeError('Socket ID should %s' % text)

        if custom_data:
            custom_data = json.dumps(custom_data)

        string_to_sign = "%s:%s" % (socket_id, channel)

        if custom_data:
            string_to_sign += ":%s" % custom_data

        signature = hmac.new(self.secret.encode('utf8'), string_to_sign.encode('utf8'), hashlib.sha256).hexdigest()

        auth = "%s:%s" % (self.key, signature)
        result = {'auth': auth}

        if custom_data:
            result['channel_data'] = custom_data

        return result

    def validate_webhook(self, key, signature, body):
        """Used to validate incoming webhook messages. When used it guarantees
        that the sender is Pusher and not someone else impersonating it.

        :param key: key used to sign the body
        :param signature: signature that was given with the body
        :param body: content that needs to be verified
        """
        if not isinstance(key, six.text_type):
            raise TypeError('key should be %s' % text)

        if not isinstance(signature, six.text_type):
            raise TypeError('signature should be %s' % text)

        if not isinstance(body, six.text_type):
            raise TypeError('body should be %s' % text)

        if key != self.key:
            return None

        generated_signature = six.text_type(hmac.new(self.secret.encode('utf8'), body.encode('utf8'), hashlib.sha256).hexdigest())

        if not compare_digest(generated_signature, signature):
            return None

        try:
            body_data = json.loads(body)
        except ValueError:
            return None

        time_ms = body_data.get('time_ms')
        if not time_ms:
            return None

        print(abs(time.time()*1000 - time_ms))
        if abs(time.time()*1000 - time_ms) > 300000:
            return None

        return body_data



# -*- coding: utf-8 -*-

from __future__ import (print_function, unicode_literals, absolute_import,
                        division)
from pusher.util import GET, POST

import copy
import hashlib
import hmac
import json
import six
import time

def make_query_string(params):
    return '&'.join(map('='.join, sorted(params.items(), key=lambda x: x[0])))

class Request(object):
    """Represents the request to be made to the Pusher API.

    An instance of that object is passed to the backend's send_request method
    for each request.

    :param config: an instance of pusher.Config
    :param method: HTTP method as a string
    :param path: The target path on the destination host
    :param params: Query params or body depending on the method
    """
    def __init__(self, config, method, path, params={}):
        self.config = config
        self.method = method
        self.path = path
        self.params = copy.copy(params)
        if method == POST:
            self.body = six.text_type(json.dumps(params)).encode('utf8')
            self.query_params = {}
        elif method == GET:
            self.body = bytes()
            self.query_params = params
        else:
            raise NotImplementedError("Only GET and POST supported")
        self._generate_auth()

    def _generate_auth(self):
        self.body_md5 = hashlib.md5(self.body).hexdigest()
        self.query_params.update({
            'auth_key': self.config.key,
            'body_md5': six.text_type(self.body_md5),
            'auth_version': '1.0',
            'auth_timestamp': '%.0f' % time.time()
        })

        auth_string = '\n'.join([
            self.method,
            self.path,
            make_query_string(self.query_params)
        ])

        secret = self.config.secret.encode('utf8')
        message = auth_string.encode('utf8')

        self.query_params['auth_signature'] = six.text_type(hmac.new(secret, message, hashlib.sha256).hexdigest())

    @property
    def query_string(self):
        return make_query_string(self.query_params)

    @property
    def signed_path(self):
        return "%s?%s" % (self.path, self.query_string)

    @property
    def url(self):
        return "%s://%s:%s%s" % (self.config.scheme, self.config.host, self.config.port, self.signed_path)

    @property
    def headers(self):
        if self.method == POST:
            return {"Content-Type": "application/json"}
        else:
            return {}



# -*- coding: utf-8 -*-

from __future__ import (print_function, unicode_literals, absolute_import,
                        division)
from pusher.util import PusherError, process_response
from six.moves import http_client

import socket
import ssl
import sys

class SynchronousBackend(object):
    """Adapter for the standard-library http client.

    :param config: an instance of pusher.Config
    :param timeout: configurable timeout for the TCP connection
    """
    def __init__(self, config, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
        self.config = config
        self.timeout = timeout
        if config.ssl:
            if sys.version_info < (3,4):
                raise NotImplementedError("SSL requires python >= 3.4, earlier versions don't support certificate validation")

            ctx = ssl.create_default_context()
            self.http = http_client.HTTPSConnection(self.config.host, self.config.port, timeout=self.timeout, context=ctx)
        else:
            self.http = http_client.HTTPConnection(self.config.host, self.config.port, timeout=self.timeout)

    def send_request(self, request):
        try:
            self.http.request(request.method, request.signed_path, request.body, {"Content-Type": "application/json"})
            resp = self.http.getresponse()
            body = resp.read().decode('utf8')
        except http_client.HTTPException as e:
            raise PusherError(repr(e))

        return process_response(resp.status, body)
